// by Alexander Ilyin
// errors handling (reporting, and not only)
// original belongs at "Drag Watcher" project

#include "error.h"
#ifdef WIN32
#include <windows.h>
#else
#include <stdarg.h>
#endif
//#include <stdlib.h>
#include <string.h>
#include <stdio.h>
//#include "config.h"
#include "types.h"


#define PROJECT_ERROR PROJECT_NAME" - error"

static char error_exit_message[1024] = {0};


#if ERROR_WITH_LOGFILE
#define ERROR_LOGFILE_NAME "error_log.txt"

void Error_Log_Print (const char* str)
{
#ifdef WIN32
	HANDLE h_log_file;
	uint string_length;
	BOOL B_result;
	DWORD bytes_written = 0;
	bool is_line_feed;
	bool is_carriage_return;
	bool was_carriage_return;
	char c;
	uint i;

	was_carriage_return = 0;
	string_length = strlen (str);
	h_log_file = CreateFile (ERROR_LOGFILE_NAME, GENERIC_WRITE, FILE_SHARE_READ, 0, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
	if (h_log_file == INVALID_HANDLE_VALUE) 
		return;
	DWORD D_result = SetFilePointer (h_log_file, 0, NULL, FILE_END);
// :P
//	if (D_result == 0xffffffff)
//		Error ("Error_Log_Print () :  'SetFilePointer ()' fails");
#if 0
	B_result = WriteFile (h_log_file, str, string_length, &bytes_written, 0);
#else
// convert '\r' and '\n' endings to the proper for ASCII "\r\n" 
// (yes, the word "proper" here has a very relative/doubtful meaning; but the basic point is that the ASCII doesn't have "line end" or "new line" or "next line";
// it has only a printer's "line feed" 0x0a and "carriage return" 0x0d (why have I write such a crappy notice, every time when dealing with this stuff :P);
// the Control Functions C0 set is in ISO/IEC 6429 and ECMA-48)
// some texts have "\r\r\n" and "\r\n\n" inplace of a single line end, and not of two, so they definitely are the bugs, and should not be treaten as a single line end,
// and will be treaten as a two line ends of an inconsistent line ended text.
	for (i = 0; i < string_length; i++)
	{
		c = str[i];
		is_line_feed = c == '\n';
		is_carriage_return = c == '\r';
// woah how complex :) ; not really complex, if remember that it let to avoid of three conditional 'WriteFile ()'
// so it is :  write "\r\n" if it is '\n' but was no '\r' or write "\r\n" if it is '\r'
		if (is_line_feed && was_carriage_return)
// already printed
			goto continue_loop;
		if (is_carriage_return || is_line_feed)
		{
			B_result = WriteFile (h_log_file, "\r\n", 2, &bytes_written, 0);
		}
		else
			B_result = WriteFile (h_log_file, &c, 1, &bytes_written, 0);
continue_loop:
		was_carriage_return = is_carriage_return;
	}
#endif

// finish the line, if it isn't finished
	if (string_length)
		if (! is_line_feed && ! is_carriage_return)
			B_result = WriteFile (h_log_file, "\r\n", 2, &bytes_written, 0);

// put a separator, for good readability
	const char* separator;
	separator = "\r\n-----\r\n\r\n";
	string_length = strlen (separator);
	B_result = WriteFile (h_log_file, separator, string_length, &bytes_written, 0);

	CloseHandle (h_log_file);
#endif
#ifdef __unix__
#endif
}

void Error_Log_Printf (const char* format, ...)
{
	va_list arglist;

	va_start(arglist, format);
	vsprintf (error_exit_message, format, arglist);
	va_end(arglist);

	Error_Log_Print (error_exit_message);
}
#endif

// no need to restrict it to 'static', and no need to declare it :)
void ErrorReport_VA (const char* format, va_list arglist)
{
//	strcpy (error_exit_message,"\nError: ");
//	va_start(arglist, format);
//	vsprintf (error_exit_message + strlen(error_exit_message), format, arglist);
	vsprintf (error_exit_message, format, arglist);
//	va_end(arglist);

#ifndef ERROR_TO_CONSOLE
#error must be defined in config
#endif
#if ERROR_TO_CONSOLE
// *begin from '\n', because there could be an unfinished string
	printf ("\n%s\n", error_exit_message);
#endif

#ifndef ERROR_TO_MBOX
#error must be defined in config
#endif
#if ERROR_TO_MBOX
	MessageBox (0, error_exit_message, PROJECT_ERROR, MB_OK);
#endif
}

// yet have to leave this basic function, because it will always be necessary, because not always is necessary to print the function name, where a fail occurs
void Error (const char* format, ...)
{
	va_list arglist;

	va_start(arglist, format);
	ErrorReport_VA (format, arglist);
	va_end(arglist);

// *will call all declared by 'atexit()'
	exit (-1);
}

// and this one is for use in a macros
// if 'do_exit' is 0, then will not exit and the process will continue
void _Error (bool do_exit, const char* format, ...)
{
	va_list arglist;

	va_start(arglist, format);
	ErrorReport_VA (format, arglist);
	va_end(arglist);

	if (do_exit)
		exit (-1);
}

// *the return values is not used; just to satisfy a newer compilers, when doing '?:'
bool _AssertionFailed (bool do_exit, const char* expr_text, const char* filename, int linenum, const char* funcname)
{
// *these error are reports about an abnormal program behaviour; they have no any need to be on a single line
//	_Error (do_exit, "Assertion failed: '%s', file \"%s\", line %d, %s()", expr_text, filename, linenum, funcname);
	_Error (do_exit, "Assertion failed:\n'%s'\nfile \"%s\", line %d, %s()", expr_text, filename, linenum, funcname);
	return 0;
}

