#pragma once

#include "config.h"


// *MSVC 6.0 does not know '__FUNCTION__'
// instead of writing two kind of functions, it easier can be hacked
// '#if defined(__FUNCTION__)' always fails on GCC; though who cares (?) - on MSVC it works properly
// also, if on GCC define '__FUNCTION__', then it will overload the regular one
//// *MSDN offers '_MSC_VER < 1300' in "aa449711" (which is MSVS 7.0)
//// '__FUNCTION__' is in the "Product Changes: Visual C++ .NET 2002" list (2byy0fh6)
// '__FUNCTION__' is introduced in MSVC 2002 :  http://msdn.microsoft.com/en-us/library/2byy0fh6%28v=vs.80%29.aspx
//#if _MSC_VER < 1300
#if ! defined(__FUNCTION__)
#define __FUNCTION__ "---"
#endif

#if ! defined (ERROR_WITH_LOGFILE)
#define ERROR_WITH_LOGFILE 0
#endif

#if ERROR_WITH_LOGFILE
void Error_Log_Print (const char* str);
void Error_Log_Printf (const char* format, ...);
#endif
void Error (const char* format, ...);
void _Error (bool do_exit, const char* format, ...);

// a program should not call a function, just to check a single condition; so this function gets called when the condition fails
// MSVC 6.0 and 8.0 have no problem with the void return type inside of '?:' operator, but GCC 4.6.0 does a warning "... of type 'void' ..."
bool _AssertionFailed (bool do_exit, const char* expr_text, const char* filename, int linenum, const char* funcname);

// these macros are so basic and common, so can shorten them
// *looks like the readability is making a bit worse :/
// *"Debug break" is a senseless report :  need break - set a breakpoint, need error report - use ERROR_THROW
//#define Brk ERROR_THROW
// *it is not obvious, that 'WARNING' would do an error report, and then do not exit, and not a warning report; however, it may still appear usable in this way
#define WARNING ER
#define W WARNING
#define WT ERT

// 'do, while' adds 'xor, test, jnz' overhead for MSVC 6.0, in the speed optimizing and nonoptimizing modes
#define A(_expression_) ((_expression_) ? 0 : _AssertionFailed (1, #_expression_, __FILE__, __LINE__, __FUNCTION__))

// *MSVCs up to 2005 do not compile ellipsis
// *Variadic Macros support has been introduced in MSVC 2005
#if _MSC_VER >= 1400
// ERROR
// eeeh, undefine it in "wingdi.h" - it is nonsense to use such important define for some garbage
//#undef ERROR
#define E(_format, ...) do { _Error (1, __FUNCTION__" () :  "_format, __VA_ARGS__); } while (0)
// ERROR_REPORT
#define ER(_format, ...) do { _Error (0, __FUNCTION__" () :  "_format, __VA_ARGS__); } while (0)
// don't write A(0) or assert report - use these instead
// ERROR_THROW
#define ET do { _Error (1, __FUNCTION__" () :  <unconditional error>.\nfile \"%s\", line %d.", __FILE__, __LINE__); } while (0)
// ERROR_REPORT_THROW
#define ERT do { _Error (0, __FUNCTION__" () :  <unconditional error>.\nfile \"%s\", line %d.", __FILE__, __LINE__); } while (0)
#else
// hack this MSVC 6.0 problem
// MSVC 6.0 doesn't know even '__FUNCTION__', which was introduced in MSVC 2002
#include <stdio.h>
//#include <stdlib.h>
//#include <varargs.h>
#include <stdarg.h>
static const char* error_filename_ = 0;
static int error_line_ = 0;
// *'__inline' is not enough, and should be 'static'; otherwise gives wrong filename and line sometimes
//__inline void Error_ (const char* format, ...)
static void Error_ (const char* format, ...)
{
	static char buffer[1000];
	va_list argptr;

	va_start (argptr, format);
	_vsnprintf (buffer, sizeof(buffer) - 1, format, argptr);
	va_end (argptr);
	_Error (1, "file \"%s\", line %u;\n%s", error_filename_, error_line_, buffer);
}
//__inline void ErrorReport_ (const char* format, ...)
static void ErrorReport_ (const char* format, ...)
{
	static char buffer[1000];
	va_list argptr;

	va_start (argptr, format);
	_vsnprintf (buffer, sizeof(buffer) - 1, format, argptr);
	va_end (argptr);
	_Error (0, "file \"%s\", line %u;\n%s", error_filename_, error_line_, buffer);
}
// horrible, I know
// *note the space between macro name and everything other (the left parenthesis)
#define E (error_filename_ = __FILE__, error_line_ = __LINE__), Error_
#define ER (error_filename_ = __FILE__, error_line_ = __LINE__), ErrorReport_
#define ET do { _Error (1, "unconditional error.\nfile \"%s\", line %d.", __FILE__, __LINE__); } while (0)
#define ERT do { _Error (0,  "unconditional error.\nfile \"%s\", line %d.", __FILE__, __LINE__); } while (0)
#endif

// for align purposes, 'value' goes last, and not the first
#define ASSERT_RANGE(_minimum, _maximum, _value) A((_value) >= (_minimum) && (_value) < (_maximum))

// *'ER()' here is not useful; it is better to use specific format for an assertions
// *I'll prepend underscore at those which aren't for a public use
#define _REPORT_FAILED_ASSERTION(_assertion) \
do { \
	ER("failed assertion :  '"#_assertion"'"); \
} while (0)
// assertion with just report, and not a dead loop
// ASSERT_REPORT
#define AR(_assertion) \
do { \
	if (! (_assertion)) \
		_REPORT_FAILED_ASSERTION(_assertion); \
} while (0)

// ow, it is not hackable the way that the 'ER()' is
// so MSVCs below 2005 are being flushed down
#if _MSC_VER >= 1400
// ERROR_REPORT_RET_VOID
#define ERR_V(_format, ...) \
do { \
	ER(_format, __VA_ARGS__); \
	return; \
} while (0)
// ERROR_REPORT_RET_0
#define ERR_0(_format, ...) \
do { \
	ER(_format, __VA_ARGS__); \
	return 0; \
} while (0)
#endif
// ERROR_REPORT_THROW_RET_VOID
#define ERTR_V \
do { \
	ERT; \
	return; \
} while (0)
// ERROR_REPORT_THROW_RET_0
#define ERTR_0 \
do { \
	ERT; \
	return 0; \
} while (0)

#define _REPORT_FAILED_ASSERT_RET_VOID(_assertion) \
do { \
	_REPORT_FAILED_ASSERTION(_assertion); \
	return; \
} while (0)
// ASSERT_REPORT_RET_VOID
#define ARR_V(_assertion) \
do { \
	if (! (_assertion)) \
		_REPORT_FAILED_ASSERT_RET_VOID(_assertion); \
} while (0)

#define _REPORT_FAILED_ASSERT_RET_0(_assertion) \
do { \
	_REPORT_FAILED_ASSERTION(_assertion); \
	return 0; \
} while (0)
// ASSERT_REPORT_RET_0
#define ARR_0(_assertion) \
do { \
	if (! (_assertion)) \
		_REPORT_FAILED_ASSERT_RET_0(_assertion); \
} while (0)

