// by Alexander Ilyin
//// code pieces do belong to Parallax Software Corporation and Outrage Entertainment Inc.
// written looking by code which is by Parallax Software Corporation
// HOG files

#include "hog.h"
#include "error.h"
#include <string>
using namespace std;
//#include <ansiapi.h>


// these macros are 'W2A()' and 'A2W()' from <ansiapi.h> of MSVC 6.0, but that file or these macros are not present in MSVC 8.0
#define WTOA(_w, _a, _cb) WideCharToMultiByte (CP_ACP, 0, _w, -1, _a, _cb, 0,0)
#define ATOW(_a, _w, _cb) MultiByteToWideChar (CP_ACP, 0, _a, -1, _w, _cb)


//// eheh, I dislike to use a constructor here, but this is the mostly convenient way :/
#if 0
Hog::Hog ()
{
// *it also actually zeroes the content of 'm_file'
	memset (this, 0, sizeof(*this));
}
#endif

void Hog::ConvertError ()
{
	switch (m_file.m_error)
	{
	case FILE_ERR__NONE:
//		return 0;
// this function should not be called if there was no error
// (theoretically it could be called in any case, but for now I made it to be called only in the case of an error)
		ET;
	case FILE_ERR__OUT_OF_MEMORY:
		m_error = PACK_ERR__OUT_OF_MEMORY;
		break;
	case FILE_ERR__UNSPECIFIED_OPEN_ERROR:
		m_error = PACK_ERR__UNSPECIFIED_OPEN_ERROR;
		break;
	case FILE_ERR__NO_FILE:
		m_error = PACK_ERR__NO_FILE;
		break;
	case FILE_ERR__NO_ACCESS:
		m_error = PACK_ERR__NO_ACCESS;
		break;
	case FILE_ERR__CLOSE_ERROR:
		m_error = PACK_ERR__CLOSE_ERROR;
		break;
	case FILE_ERR__READ_FAULT:
		m_error = PACK_ERR__READ_FAULT;
		break;
	case FILE_ERR__INCOMPLETE_READ:
//// should be dispatched
//		ET;
		m_error = PACK_ERR__READ_FAULT;
		break;
	case FILE_ERR__WRITE_FAULT:
		m_error = PACK_ERR__WRITE_FAULT;
		break;
	case FILE_ERR__DISK_FULL:
		m_error = PACK_ERR__DISK_FULL;
	default:
		ET;
	}
//	return 0;
}

bool Hog::Open (bool write, bool widechar, const void* filepath)
{
	bool r;

//// *it also actually zeroes the content of 'm_file'
//	memset (this, 0, sizeof(*this));
//	A(! m_is_opened);
//	m_write = write;

	m_error = PACK_ERR__UNSPECIFIED_OPEN_ERROR;

	r = m_file.Open (write, widechar, filepath);
	if (! r)
		goto file_error;

	if (! write)
	{
		r = m_file.GetTime (m_packfile_filetime);
		if (! r)
			goto file_error;
	}

//	m_is_opened = 1;
	m_error = PACK_ERR__NONE;
	return 1;

file_error:
	ConvertError ();
	return 0;
}

bool Hog::Close ()
{
	bool r;

//	A(m_is_opened);
	m_error = PACK_ERR__CLOSE_ERROR;
	r = m_file.Close ();
	if (! r)
	{
		ConvertError ();
		return 0;
	}
//	m_is_opened = 0;
	m_error = PACK_ERR__NONE;
	return 1;
}

bool Hog::ReadHdr ()
{
	bool r;
	char buf[3];

//	A(m_is_opened);
	m_error = PACK_ERR__READ_FAULT;
	r = FRD(buf);
	if (! r)
	{
		if (m_file.m_error == FILE_ERR__INCOMPLETE_READ)
			goto bad_header;
		ConvertError ();
		return 0;
	}
	if (memcmp (buf, "DHF", 3))
		goto bad_header;
// and that's all; the header has nothing more except three characters "DHF" in it :/
	m_hog2 = 0;
	m_error = 0;
	return 1;

bad_header:
	m_error = PACK_ERR__BAD_HEADER;
	return 0;
}

bool Hog::WriteHdr ()
{
	bool r;
	static char id_hog[4] = "DHF";
	static char id_hog2[5] = "HOG2";
	u32 len;

//	A(m_is_opened);
	m_error = PACK_ERR__WRITE_FAULT;
	len = m_hog2 ? 4 : 3;
	r = m_file.Write (m_hog2 ? id_hog2 : id_hog, len);
	if (! r)
	{
		ConvertError ();
		return 0;
	}
	if (! m_hog2)
		goto finish;

// *HOG2 id size is 4; header size is 0x40, but only 8 bytes of it are used; thus the entries begin from 0x44
	;

finish:
//	m_error = -1;
	m_error = 0;
	return 1;
}

//// *doesn't return 'm_error' on fail
//bool Hog::HasMoreFiles ()
bool Hog::HasMoreFiles (bool& has_more)
{
	bool r;

//	A(m_is_opened);
//	has_more = 0;
	m_error = PACK_ERR__READ_FAULT;
	r = m_file.IsNotEOF (has_more);
	if (! r)
	{
		ConvertError ();
		return 0;
	}
	m_error = PACK_ERR__NONE;
	return 1;
}

// *call 'HasMoreFiles()' before that, otherwise it will report about bad header
bool Hog::ReadEntry ()
{
	bool r;

//	A(m_is_opened);
//	A(! m_write);
	A(! m_hog2);
	m_error = PACK_ERR__READ_FAULT;

	r = m_file.Read (m_name, HOG_ENTRY_FILE_NAME_BUF_SIZE);
	if (! r)
	{
		if (m_file.m_error == FILE_ERR__INCOMPLETE_READ)
			goto bad_header;
		goto file_error;
	}
// ensure 0 ending
	m_name[HOG_ENTRY_FILE_NAME_LEN_MAX] = 0;

	r = FRD(m_len);
	if (! r)
	{
		if (m_file.m_error == FILE_ERR__INCOMPLETE_READ)
			goto bad_header;
		goto file_error;
	}

	m_hdr_len = HOG_ENTRY_FILE_NAME_BUF_SIZE + 4;
	m_error = PACK_ERR__NONE;

	return 1;

file_error:
	ConvertError ();
	return 0;

bad_header:
	m_error = PACK_ERR__BAD_HEADER;
	return 0;
}

bool Hog::WriteEntry ()
{
	bool r;

//	A(m_write);
	A(! m_hog2);
//	m_error = PACK_ERR__WRITE_FAULT;
	m_error = -1;

	r = m_file.Write (m_name, HOG_ENTRY_FILE_NAME_BUF_SIZE);
	if (! r)
		goto file_error;

	r = FWR(m_len);
	if (! r)
		goto file_error;

	return 1;

file_error:
	ConvertError ();
	return 0;
}

bool Hog::SkipFile ()
{
	bool r;

//	A(m_is_opened);
	m_error = -1;

	r = m_file.SetPointer (1, m_len);
	if (! r)
	{
		ConvertError ();
		return 0;
	}
//	m_len = 0;
	m_error = PACK_ERR__NONE;
	return 1;
}

// these functions became a bit extra
#if 0
bool Hog::ReadToMem ()
{
	bool r;

//	A(! m_write);
	m_error = -1;

	r = m_file.Read (m_mem, m_len);
#if 1
	if (! r)
		return 0;
#else
	if (! r)
	{
		ConvertError ();
		return 0;
	}
	m_error = PACK_ERR__NONE;
#endif

	return 1;
}

bool Hog::WriteFromMem ()
{
	bool r;

//	A(m_write);
	m_error = -1;

	r = m_file.Write (m_mem, m_len);
	if (! r)
		return 0;

	return 1;
}
#endif

bool Hog::CopyToFile_Mallocated (bool widechar, const void* filepath)
{
	bool r;
	File fo;
//	File fo = {0};

	memset (&fo, 0, sizeof(fo));

	r = fo.Open (1, widechar, filepath);
	if (! r)
		goto file_error;
//		return 0;

	r = m_file.Read (m_mem, m_len);
	if (! r)
		goto file_error;

	r = fo.Write (m_mem, m_len);
	if (! r)
		goto file_error;

// change the time when the file was modified
	r = fo.SetTime (m_packfile_filetime);
	if (! r)
		goto file_error;

	m_error = PACK_ERR__NONE;

finish:
	if (fo.m_is_opened)
	{
		r = fo.Close ();
		if (! r)
			if (! m_error)
// **err, it isn't for this file
				ConvertError ();
	}
//	return 1;
	return ! m_error;

file_error:
	ConvertError ();
	A(m_error);
//	return 0;
	goto finish;
}

// copy data from under the current pointer, and of 'm_len' length into a file
bool Hog::CopyToFile (bool widechar, const void* filepath)
{
	bool r;

//	m_error = -1;

//	A(! m_write);
	A(! m_mem);

	m_mem = malloc (m_len);
	if (! m_mem)
	{
		m_error = PACK_ERR__OUT_OF_MEMORY;
		return 0;
	}

	r = CopyToFile_Mallocated (widechar, filepath);
	if (! r)
	{
		A(m_error);
		goto finish;
	}

//	m_error = 0;

finish:
	free (m_mem);
	m_mem = 0;
//	return 1;
	return ! m_error;
}

bool Hog::SetVersion (bool is_hog2)
{
	m_hog2 = is_hog2;
	m_filename_len_max = is_hog2 ? HOG2_ENTRY_FILE_NAME_LEN_MAX : HOG_ENTRY_FILE_NAME_LEN_MAX;
	m_filename_buf_size = is_hog2 ? HOG2_ENTRY_FILE_NAME_BUF_SIZE : HOG_ENTRY_FILE_NAME_BUF_SIZE;
	return 1;
}

// the characters of it must be in 0x20-0x80 range
// (there is no such limitation in the actual HOG, but just I am using it for HOG identification; and such a check actually will not harm)
bool Hog::CheckFileNameForEntry ()
{
	char c;
	const char* ptr;

	ptr = m_name;
	if (! *ptr)
		return 0;
	while ((c = *ptr))
	{
//		if (! c)
//			return 1;
		if (c < 0x20)
			return 0;
		if (c >= 0x80)
			return 0;
		ptr++;
	}
	return 1;
}

bool Hog::PutFileName (bool widechar, const void* filename)
{
	bool r;
	u32 i;

	A(m_filename_len_max);
	A(m_filename_len_max < HOGx_ENTRY_FILE_NAME_BUF_SIZE_MAX);
	A(m_filename_buf_size);
	A(m_filename_buf_size <= HOGx_ENTRY_FILE_NAME_BUF_SIZE_MAX);

	m_error = PACK_ERR__FILENAME_INCOMPATIBLE;
// *simplified
	memset (m_name, 0, m_filename_buf_size);

// the name must fit in the HOG's name entry
	if (widechar)
	{
// get required buffer length
		i = WTOA((wchar_t*)filename, 0, 0);
		if (i > m_filename_buf_size)
			return 0;
		i = WTOA((wchar_t*)filename, m_name, m_filename_buf_size);
// should not fail
		A(i);
	}
	else
	{
//		strncpy (m_name, filename, HOG_ENTRY_FILE_NAME_BUF_SIZE);
		i = strlen ((char*)filename);
		if (i > m_filename_len_max)
			return 0;
		strcpy (m_name, (char*)filename);
	}

	r = CheckFileNameForEntry ();
	if (! r)
		return 0;

// *unset
	m_error = -1;

	return 1;
}

#if 0
bool Hog::CopyFromFile_Opened (File& fi)
{
	bool r;

	r = fi.GetSize ();
	if (! r)
		goto file_error;
	if (fi.m_len >= HOG_ENTRY_FILE_SIZE_MAX)
	{
		m_error = PACK_ERR__FILE_TOO_BIG;
//		goto finish;
		return 0;
	}
	m_len = fi.m_len;

	return 1;

file_error:
	ConvertError ();
	return 0;
}
#endif

// it is not add any convenience if split this function onto a three, including each other; that instead would add a repeated pieces of code, containing 'ConvertError()' calls
bool Hog::CopyFromFile (bool widechar, const void* path, const void* filename)
{
	bool r;
	File fi;

//	A(m_write);
	A(! m_mem);
	m_error = -1;

	r = PutFileName (widechar, filename);
	if (! r)
		return 0;

	memset (&fi, 0, sizeof(fi));

	if (widechar)
	{
		wstring ws;
		ws = (wchar_t*)path;
		ws += (wchar_t*)filename;
		r = fi.Open (0, widechar, ws.c_str());
	}
	else
	{
		string s;
		s = (char*)path;
		s += (char*)filename;
		r = fi.Open (0, widechar, s.c_str());
	}
	if (! r)
		goto file_error;

	r = fi.GetSize ();
	if (! r)
		goto file_error;
	if (fi.m_len >= HOG_ENTRY_FILE_SIZE_MAX)
	{
		m_error = PACK_ERR__FILE_TOO_BIG;
		goto finish;
	}
	m_len = fi.m_len;

	r = WriteEntry ();
	if (! r)
		goto finish;

	m_mem = malloc (m_len);
	if (! m_mem)
	{
		m_error = PACK_ERR__OUT_OF_MEMORY;
		goto finish;
	}

	r = fi.Read (m_mem, m_len);
	if (! r)
		goto file_error;

	r = m_file.Write (m_mem, m_len);
	if (! r)
		goto file_error;

	m_error = PACK_ERR__NONE;

finish:
	if (m_mem)
	{
		free (m_mem);
		m_mem = 0;
	}
	if (fi.m_is_opened)
	{
		r = fi.Close ();
		if (! r)
			if (! m_error)
// **err, it isn't for this file
				ConvertError ();
	}
//	return 1;
	return ! m_error;

file_error:
	ConvertError ();
	A(m_error);
//	return 0;
	goto finish;
}

