#pragma once

#include "types.h"
#include "file.h"


// *wrong, it is not a file error, but a set of errors which may occur at a pack/unpack time
// it more is like a container error
enum
{
	PACK_ERR__NONE,
	PACK_ERR__OUT_OF_MEMORY,
	PACK_ERR__UNSPECIFIED_OPEN_ERROR,
	PACK_ERR__NO_FILE,
	PACK_ERR__NO_ACCESS,
	PACK_ERR__CLOSE_ERROR,
	PACK_ERR__READ_FAULT,
	PACK_ERR__WRITE_FAULT,
	PACK_ERR__DISK_FULL,
// this isn't our file
	PACK_ERR__BAD_HEADER,
// this file is our, but we don't know this exact format
	PACK_ERR__UNSUPPORTED_FORMAT,
// fileinfo header is broken
	PACK_ERR__BAD_ENTRY,
// file data is incomplete
	PACK_ERR__BAD_FILE,
// this file cannot be put into the HOG
	PACK_ERR__FILE_TOO_BIG,
// this filename can't be put in the HOG
	PACK_ERR__FILENAME_INCOMPATIBLE,
	PACK_ERRS_NUM
};

// HOG and HOG2 :  situation is the same - the ending zero is stored inside of the HOGs
#if 0
#define HOG_ENTRY_FILE_NAME_BUF_SIZE 13
#define HOG_ENTRY_FILE_NAME_LEN_MAX (HOG_ENTRY_FILE_NAME_BUF_SIZE - 1)
#else
#define HOG_ENTRY_FILE_NAME_LEN_MAX 12
#define HOG_ENTRY_FILE_NAME_BUF_SIZE (HOG_ENTRY_FILE_NAME_LEN_MAX + 1)
#endif
// PSFILENAME_LEN
#define HOG2_ENTRY_FILE_NAME_LEN_MAX 35
#define HOG2_ENTRY_FILE_NAME_BUF_SIZE (HOG2_ENTRY_FILE_NAME_LEN_MAX + 1)
#define HOGx_ENTRY_FILE_NAME_BUF_SIZE_MAX (HOG2_ENTRY_FILE_NAME_BUF_SIZE)
#define HOG_ENTRY_FILE_SIZE_MAX 0x40000000

class Hog
{
private:
	void ConvertError ();
public:
//	Hog ();
//	bool Open (const char* filepath);
//	bool Open (const wchar_t* filepath);
//	bool Open (bool widechar, const void* filepath);
	bool Open (bool write, bool widechar, const void* filepath);
	bool Close ();
// *'ReadHeader()' conflicts at linking
	bool ReadHdr ();
	bool WriteHdr ();
	bool HasMoreFiles (bool& has_more);
	bool ReadEntry ();
private:
	bool WriteEntry ();
public:
	bool SkipFile ();
private:
#if 0
	bool ReadToMem ();
	bool WriteFromMem ();
#endif
	bool CopyToFile_Mallocated (bool widechar, const void* filepath);
public:
	bool CopyToFile (bool widechar, const void* filepath);
	bool SetVersion (bool is_hog2);
private:
	bool CheckFileNameForEntry ();
	bool PutFileName (bool widechar, const void* filename);
public:
//	bool CopyFromFile (bool widechar, const void* filepath);
	bool CopyFromFile (bool widechar, const void* path, const void* filename);

private:
#if 0
// here :  { un u; u.c = 2; cout << u.r; }   union un { char c; bool r; };
// Clang 3.2 returns 0, and GCC 4.7.0 returns 1
	u8 m_dummy;
// *use it in conjunction with 'm_dummy' in an assertions
#endif
//	bool m_is_opened;
//	HANDLE m_hfile;
	File m_file;
//	u8* m_mem;
	void* m_mem;
// *helper
//	bool m_write;
	u32 m_filename_len_max;
	u32 m_filename_buf_size;
public:
// *note, that I set it to -1, when the operation is not yet completed/finished (some error still may occur ahead; useable in a private functions)
	u32 m_error;
#if 0
// remember, what name it has
	char m_arc_name[1024];
//	WCHAR m_arc_name[1024];
#endif
	FILETIME m_packfile_filetime;
	bool m_hog2;
	char m_name[HOG_ENTRY_FILE_NAME_BUF_SIZE];
	u32 m_len;
	u32 m_hdr_len;
};

#define FRD(_obj) m_file.Read (&_obj, sizeof(_obj))
#define FWR(_obj) m_file.Write (&_obj, sizeof(_obj))

