// by Alexander Ilyin
//

#include <windows.h>
#include "wcxhead.h"
#include "error.h"
#include "hog.h"
#include <string>
using namespace std;


HINSTANCE g_hinstance = 0;
// +++ settings +++
bool is_hog2 = 0;
// --- settings ---


#if 0
template<typename T>
T* new0 ()
{
	T* obj;
	obj = new T;
	if (obj)
		memset (obj, 0, sizeof(*obj));
	return obj;
}
#endif
#if 1
template<typename T>
void new0 (T*& obj)
{
	obj = new T;
	if (obj)
		memset (obj, 0, sizeof(*obj));
}
#endif

BOOL APIENTRY DllMain (HINSTANCE hinst, DWORD reason, LPVOID reserved)
{
	switch (reason)
	{
	case DLL_PROCESS_ATTACH:
		g_hinstance = hinst;
		break;
	case DLL_PROCESS_DETACH:
		break;
	default:
		break;
	}
	return 1;
}

u32 ConvertError (Hog* hog)
{
	A(hog);
	switch (hog->m_error)
	{
	case PACK_ERR__NONE:
//		return 0;
// this function should not be called if there was no error
// (theoretically it could be called in any case, but for now I made it to be called only in the case of an error)
		ET;
	case PACK_ERR__OUT_OF_MEMORY:
		return E_NO_MEMORY;
	case PACK_ERR__UNSPECIFIED_OPEN_ERROR:
	case PACK_ERR__NO_FILE:
	case PACK_ERR__NO_ACCESS:
		return E_EOPEN;
	case PACK_ERR__CLOSE_ERROR:
		return E_ECLOSE;
	case PACK_ERR__READ_FAULT:
		return E_EREAD;
// *would be good to report about "disk full", but Total Commander doesn't offer that error
	case PACK_ERR__WRITE_FAULT:
	case PACK_ERR__DISK_FULL:
		return E_EWRITE;
	case PACK_ERR__BAD_HEADER:
		return E_BAD_ARCHIVE;
	case PACK_ERR__UNSUPPORTED_FORMAT:
		return E_UNKNOWN_FORMAT;
	case PACK_ERR__BAD_ENTRY:
	case PACK_ERR__BAD_FILE:
		return E_BAD_ARCHIVE;
	case PACK_ERR__FILE_TOO_BIG:
		return E_NOT_SUPPORTED;
	case PACK_ERR__FILENAME_INCOMPATIBLE:
		return E_NOT_SUPPORTED;
	default:
		ET;
	}
	return 0;
}

// the sequence of calls is :
// OpenArchive, SetProcessDataProc, SetChangeVolProc, {ReadHeader, ProcessFile}, CloseArchive

//// hmmm, can't do such a dispatcher
//static HANDLE OpenArch (* archive_data)
static HANDLE OpenArch (int& open_result, bool widechar, void* filepath)
{
	bool r;
	Hog* hog;

	A(filepath);

//	hog = new Hog;
	new0 (hog);
// no, this is too complex for a compiler, to make a deduce by return type
// { A* a; a = t (); }   struct A {};   template<typename T> T* t () { return 0; }
//	hog = new0 ();
	if (! hog)
	{
		open_result = E_NO_MEMORY;
		return 0;
	}

// yes, this is the mostly appropriate place for zeroing, but ... :/
#if 0
// *members are not zeroed by the default constructor
// zero the values
// *it may look that this initialization must be done inside of the classes, but that isn't so correct;
// for example :  what about the included 'm_file' ? - should it be zeroed inside of that class, or in the class which contains it ?
// '*obj = {0}' will not work anyway, since there are a private members
//	*hog = {0};
	memset (hog, 0, sizeof(*hog));
#endif

// *'archive_data->OpenMode' doesn't matter, because has to open the file for read anyway
	r = hog->Open (0, widechar, filepath);
	if (! r)
	{
//// hm, different from what I expected, it doesn't want to know about 'm_err_no_file' and 'm_err_no_access'
//		open_result = E_EOPEN;
		open_result = ConvertError (hog);
		goto free_obj;
	}

	r = hog->ReadHdr ();
	if (! r)
	{
		open_result = ConvertError (hog);
		r = hog->Close ();
		goto free_obj;
	}

	A(! hog->m_error);

	return hog;

free_obj:
	delete hog;
	return 0;
}

// *further 'ReadHeader()' will be called, if from here return 0 error in 'archive_data->OpenResult'
// and if not implement the ANSI functions, then this plugin will not work on older commanders
HANDLE __stdcall OpenArchive (tOpenArchiveData* archive_data)
{
	A(archive_data);
	return OpenArch (archive_data->OpenResult, 0, archive_data->ArcName);
}

// *widechar version of these functions requires Total Commander 7.5
HANDLE __stdcall OpenArchiveW (tOpenArchiveDataW* archive_data)
{
	A(archive_data);
	return OpenArch (archive_data->OpenResult, 1, archive_data->ArcName);
}

int __stdcall CloseArchive (HANDLE h_arc_data)
{
	bool r;
	Hog* hog;

	A(h_arc_data);
	hog = (Hog*)h_arc_data;
	r = hog->Close ();
	delete hog;
	if (! r)
	{
//		hog->OpenResult = E_ECLOSE;
		return E_ECLOSE;
	}
	return 0;
}

// *for this HOG project, it is more convenient to use ANSI version of 'ReadHeader()'
int __stdcall ReadHeader (HANDLE h_arc_data, tHeaderData* header_data)
{
	bool r;
	Hog* hog;
	bool has_more;
	u32 i;

#if 1
	A(h_arc_data);
#else
	if (! h_arc_data)
//		return 0;
		return E_BAD_DATA;
#endif
	A(header_data);
	hog = (Hog*)h_arc_data;
//	A(hog->);
	r = hog->HasMoreFiles (has_more);
	if (! r)
	{
		i = ConvertError (hog);
		return i;
//		return E_EREAD;
	}
	if (! has_more)
	{
		return E_END_ARCHIVE;
	}

	r = hog->ReadEntry ();
	if (! r)
	{
		i = ConvertError (hog);
		return i;
	}

	header_data->PackSize = hog->m_hdr_len + hog->m_len;
	header_data->UnpSize = hog->m_len;
// *this time isn't stored in the pack, so have to replace it with something
//	header_data->FileTime = hog->m_packfile_filetime;
	{
		FILETIME ft;
		SYSTEMTIME st;
		i = FileTimeToLocalFileTime (&hog->m_packfile_filetime, &ft);
		i = FileTimeToSystemTime (&ft, &st);
		header_data->FileTime = (st.wYear - 1980) << 25 | st.wMonth << 21 | st.wDay << 16 | st.wHour << 11 | st.wMinute << 5 | st.wSecond / 2;
	}
	memcpy (header_data->FileName, hog->m_name, 12);
// *even though it is stored in a HOG file as 0, it still should ensured to be 0, instead of just being copied from there
	header_data->FileName[12] = 0;

// is this necessary ?  for what ?
//	strcpy (header_data->ArcName, hog->m_arc_name);

	A(! hog->m_error);

	return 0;
}

//static bool ProcessFile_ (HANDLE h_arc_data, int operation, bool widechar, void* filepath)
static int ProcessFile_ (HANDLE h_arc_data, int operation, bool widechar, void* filepath)
{
	bool r;
	Hog* hog;
	u32 i;

	A(h_arc_data);
	hog = (Hog*)h_arc_data;
	switch (operation)
	{
	case PK_SKIP:
	case PK_TEST:
		r = hog->SkipFile ();
		if (! r)
			goto error;
		break;
	case PK_EXTRACT:
		r = hog->CopyToFile (widechar, filepath);
		if (! r)
			goto error;
		break;
	}

	A(! hog->m_error);

	return 0;

error:
	i = ConvertError (hog);
	return i;
}

// *XZ plugin is by Ghisler, and 'dest_path' argument in it is just ignored
int __stdcall ProcessFile (HANDLE h_arc_data, int operation, LPSTR dest_path, LPSTR dest_name)
{
	return ProcessFile_ (h_arc_data, operation, 0, dest_name);
}

int __stdcall ProcessFileW (HANDLE h_arc_data, int operation, wchar_t* dest_path, wchar_t* dest_name)
{
	return ProcessFile_ (h_arc_data, operation, 1, dest_name);
}

// note, that Total Commander calls this function only at the moment when a plugin is being installed, but later it is not being called anymore
// thus you may not get some expected behaviour, after trying to change this
// and also even in Total Commander 8.01 it isn't easy to remove a registered plugin from the list of plugins; to do that, open "WINCMD.INI" (which may be in a hidden folder), and remove the line with your plugin in [PackerPlugins] section
// otherwise you also may remove the number in that string and comma after it; this will make Total Commander to call this function again at its next start
int __stdcall GetPackerCaps ()
{
// we don't support repacking, thus PK_CAPS_MODIFY and PK_CAPS_DELETE are off
// *PK_CAPS_SEARCHTEXT is necessary for compare directories by content (otherwise there will be red question marks as the result)
	return PK_CAPS_NEW | PK_CAPS_MULTIPLE | PK_CAPS_OPTIONS | PK_CAPS_BY_CONTENT | PK_CAPS_SEARCHTEXT;
//	return PK_CAPS_NEW | PK_CAPS_MODIFY | PK_CAPS_MULTIPLE | PK_CAPS_DELETE | PK_CAPS_OPTIONS | PK_CAPS_MEMPACK	| PK_CAPS_BY_CONTENT | PK_CAPS_SEARCHTEXT;
}

void __stdcall SetChangeVolProc (HANDLE h_arc_data, tChangeVolProc change_vol_proc)
{
	A(h_arc_data);
}

void __stdcall SetChangeVolProcW (HANDLE h_arc_data, tChangeVolProcW change_vol_proc)
{
	A(h_arc_data);
}

//// *no need; files in HOG are always small
// *no need for a while
void __stdcall SetProcessDataProc (HANDLE h_arc_data, tProcessDataProc process_data_proc)
{
	A(h_arc_data);
}

void __stdcall SetProcessDataProcW (HANDLE h_arc_data, tProcessDataProcW process_data_proc)
{
	A(h_arc_data);
}

bool CanYouHandleThisFile_Opened (Hog* hog)
{
	bool r;
	bool has_more;
	u32 i;

// file of 0 size isn't valid
	r = hog->ReadHdr ();
	if (! r)
	{
		return 0;
	}

//// if the remained file size is 0, then it is ok (no files inside of it)
	r = hog->HasMoreFiles (has_more);
	if (! r)
	{
		return 0;
	}
// no files inside of it -  it is allowed
	if (! has_more)
		return 1;

// try one entry
	r = hog->ReadEntry ();
	if (! r)
	{
		return 0;
	}

// now, do some checks, which anyway will get only a result of some probability
// the length must be some meaningful
	if (hog->m_len >= HOG_ENTRY_FILE_SIZE_MAX)
		return 0;
	for (i = 0; i < HOG_ENTRY_FILE_NAME_BUF_SIZE; i++)
	{
		char c;
		c = hog->m_name[i];
// 0 must be present, but not on the first character
		if (! c)
			return !! i;
		if (c < 0x20)
			return 0;
		if (c >= 0x80)
			return 0;
	}

	return 0;
}

bool CanYouHandleThisFile_Hog (Hog* hog, bool widechar, void* filepath)
{
	bool r;
	bool result;

	r = hog->Open (0, widechar, filepath);
	if (! r)
	{
		return 0;
	}

	result = CanYouHandleThisFile_Opened (hog);

	r = hog->Close ();
	if (! r)
	{
		return 0;
	}

	return result;
}

// *for Ctrl+PgDn
static bool CanYouHandleThisFile (bool widechar, void* filepath)
{
	bool r;
//	bool result;
	Hog* hog;

	A(filepath);

//	hog = new Hog;
	new0 (hog);
	if (! hog)
	{
		return 0;
	}

	r = CanYouHandleThisFile_Hog (hog, widechar, filepath);

	delete hog;

	return r;
}

BOOL __stdcall CanYouHandleThisFile (LPSTR filepath)
{
	return CanYouHandleThisFile (0, filepath);
}

BOOL __stdcall CanYouHandleThisFileW (wchar_t* filepath)
{
	return CanYouHandleThisFile (1, filepath);
}

// *not used by "Synchronize Dirs...", even by Total Commander 8.01
#if 0
int __stdcall StartMemPack (int options, char* filepath)
{
	ET;

	return 0;
}

int __stdcall PackToMem (int h_mem_pack, char* buf_in, int len_in, int* taken, char* buf_out, int len_out, int* written, int seek_by)
{
	ET;

	return 0;
}

int __stdcall DoneMemPack (int h_mem_pack)
{
	ET;

	return 0;
}
#endif

#if 0
// the characters of it must be in 0x20-0x80 range
// (there is no such limitation in the actual HOG, but just I am using for HOG identification; and such check actually will not harm)
bool CheckNameForHogEntryFile (const char* filename)
{
	char c;

	if (! *filename)
		return 0;
	while ((c = *filename))
	{
//		if (! c)
//			return 1;
		if (c < 0x20)
			return 0;
		if (c >= 0x80)
			return 0;
		filename++;
	}
	return 1;
}
#endif

template<typename T>
//int TPackFiles_OutputOpened (Hog* hog, bool widechar, T* add_list)
//int TPackFiles_OutputOpened (Hog* hog, bool widechar, T* src_path, T* add_list)
bool TPackFiles_OutputOpened (Hog* hog, bool widechar, T* src_path, T* add_list)
{
	bool r;
//	u32 i;
//	char filename[HOGx_ENTRY_FILE_NAME_BUF_SIZE_MAX];

	r = hog->SetVersion (is_hog2);
	if (! r)
		return 0;

	r = hog->WriteHdr ();
	if (! r)
		return 0;

	while (*add_list)
	{
		r = hog->CopyFromFile (widechar, src_path, add_list);
		if (! r)
		{
			hog->m_error = E_NOT_SUPPORTED;
			return 0;
		}
// skip the name of this file
//		add_list += len;
		do
			add_list++;
		while (*add_list);
		add_list++;
	}

	A(! hog->m_error);

	return 1;
}

template<typename T>
bool TPackFiles_ToHog (Hog* hog, T* packed_file, T* src_path, T* add_list)
{
	bool r;
	bool widechar;

// static condition
	widechar = sizeof(T) == 2;
// set output format
	hog->m_hog2 = is_hog2;

	r = hog->Open (1, widechar, packed_file);
	if (! r)
		return 0;

//	result = TPackFiles_OutputOpened (hog, widechar, add_list);
//	result = TPackFiles_OutputOpened (hog, widechar, src_path, add_list);
	r = TPackFiles_OutputOpened (hog, widechar, src_path, add_list);
	if (! r)
		return 0;

	r = hog->Close ();
	if (! r)
		return 0;

	A(! hog->m_error);

	return 1;
}

template<typename T>
int TPackFiles (T* packed_file, T* sub_path, T* src_path, T* add_list, int flags)
{
	bool r;
//	bool result;
	int result;
	Hog* hog;
//	u32 i;

	A(add_list);
	if (sub_path)
		return E_NOT_SUPPORTED;
	if (flags & PK_PACK_MOVE_FILES)
		return E_NOT_SUPPORTED;
	if (flags & PK_PACK_ENCRYPT)
		return E_NOT_SUPPORTED;
	result = 0;

//	hog = new Hog;
	new0 (hog);
	if (! hog)
	{
		return E_NO_MEMORY;
	}

	r = TPackFiles_ToHog (hog, packed_file, src_path, add_list);
	if (! r)
	{
		result = ConvertError (hog);
		goto finish;
	}

	A(! hog->m_error);

finish:
	delete hog;

	return result;
}

// *it isn't called in the case of deleting a file
// but is called in the case of adding a file to an existing pack, even though PK_CAPS_MODIFY is off
int __stdcall PackFiles (LPSTR packed_file, LPSTR sub_path, LPSTR src_path, LPSTR add_list, int flags)
{
	return TPackFiles (packed_file, sub_path, src_path, add_list, flags);
}

int __stdcall PackFilesW (wchar_t* packed_file, wchar_t* sub_path, wchar_t* src_path, wchar_t* add_list, int flags)
{
	return TPackFiles (packed_file, sub_path, src_path, add_list, flags);
}

// not planned to be supported
//int __stdcall DeleteFiles (LPSTR packed_file, LPSTR delete_list)
//int __stdcall DeleteFilesW (wchar_t* packed_file, wchar_t* delete_list)

